# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing types for the Project Service."""

from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel, ValidationInfo, field_validator

from ansys.sherlock.core.types.common_types import (
    basic_list_str_validator,
    basic_str_validator,
    optional_str_validator,
)

try:
    import SherlockProjectService_pb2
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockProjectService_pb2

project_service = SherlockProjectService_pb2
thermal_map_file = project_service.ThermalMapFile
add_strain_map_request = project_service.AddStrainMapRequest


class BoardBounds:
    """Contains the properties of the board bounds."""

    def __init__(self, bounds: list[tuple[float, float]]):
        """Initialize the board bounds."""
        self.bounds = bounds
        """list[tuple[float, float]]: bounds (two tuples of the form (x, y)"""


class CsvExcelFile:
    """Contains the properties for a thermal map, CSV, or Excel file."""

    def __init__(
        self,
        header_row_count: int,
        numeric_format: str,
        reference_id_column: str,
        temperature_column: str,
        temperature_units: str,
    ):
        """Initialize the thermal map, CSV or Excel file properties."""
        self.header_row_count = header_row_count
        """int: header_row_count"""
        self.numeric_format = numeric_format
        """str: numeric_format"""
        self.reference_id_column = reference_id_column
        """str: reference_id_column"""
        self.temperature_column = temperature_column
        """str: temperature_column"""
        self.temperature_units = temperature_units
        """str: temperature_units"""


class IcepakFile:
    """Contains the properties for a thermal map Icepak file."""

    def __init__(
        self,
        temperature_offset: float,
        temperature_offset_units: str,
    ):
        """Initialize the thermal map Icepak file properties."""
        self.temperature_offset = temperature_offset
        """float: temperature_offset"""
        self.temperature_offset_units = temperature_offset_units
        """str: temperature_offset_units"""


class ImageBounds:
    """Contains the properties of the image bounds."""

    def __init__(self, image_x: float, image_y: float, height: float, width: float):
        """Initialize the image bounds properties."""
        self.image_x = image_x
        """float: x coordinate of the upper left corner"""
        self.image_y = image_y
        """float: y coordinate of the upper left corner"""
        self.height = height
        """float: height of the image"""
        self.width = width
        """float: width of the image"""


class ImageFile:
    """Contains the properties for a thermal map image file."""

    def __init__(
        self,
        board_bounds: BoardBounds,
        coordinate_units: str,
        image_bounds: ImageBounds,
        legend_bounds: "LegendBounds",
        legend_orientation: "LegendOrientation",
        max_temperature: float,
        max_temperature_units: str,
        min_temperature: float,
        min_temperature_units: str,
    ):
        """Initialize the thermal image file properties."""
        self.board_bounds = board_bounds
        """BoardBounds: board_bounds"""
        self.coordinate_units = coordinate_units
        """str: coordinate_units"""
        self.image_bounds = image_bounds
        """ImageBounds: image_bounds"""
        self.legend_bounds = legend_bounds
        """LegendBounds: legend_bounds"""
        self.legend_orientation = legend_orientation
        """LegendOrientation: legend_orientation"""
        self.max_temperature = max_temperature
        """float: max_temperature"""
        self.max_temperature_units = max_temperature_units
        """str: max_temperature_units"""
        self.min_temperature = min_temperature
        """float: min_temperature"""
        self.min_temperature_units = min_temperature_units
        """str: min_temperature_units"""


class LegendBounds:
    """Contains the properties of the legend bounds."""

    def __init__(self, legend_x: float, legend_y: float, height: float, width: float):
        """Initialize the legend bounds properties."""
        self.legend_x = legend_x
        """float: x coordinate of the upper left corner"""
        self.legend_y = legend_y
        """float: y coordinate of the upper left corner"""
        self.height = height
        """float: height of the legend"""
        self.width = width
        """float: width of the legend"""


class LegendOrientation:
    """Constants for legend orientation in the update thermal maps request."""

    __legend_orientation = thermal_map_file.ImageFile.LegendOrientation
    HORIZONTAL = __legend_orientation.Horizontal
    "Horizontal"
    VERTICAL = __legend_orientation.Vertical
    "Vertical"


class ThermalBoardSide:
    """Constants for thermal board side in the update thermal maps request."""

    BOTTOM = thermal_map_file.ThermalBoardSide.Bottom
    "Bottom"
    BOTH = thermal_map_file.ThermalBoardSide.Both
    "Both"
    TOP = thermal_map_file.ThermalBoardSide.Top
    "Top"


class ThermalMapsFileType:
    """Constants for File Type in the Update Thermal Maps request."""

    CSV = thermal_map_file.FileType.CSV
    "CSV"
    EXCEL = thermal_map_file.FileType.Excel
    "Excel"
    IMAGE = thermal_map_file.FileType.Image
    "Image"
    TMAP = thermal_map_file.FileType.TMAP
    "Icepak Thermal Map (.TMAP)"


class StrainMapsFileType:
    """Constants for File Type in the Add Strain Maps request."""

    CSV = add_strain_map_request.StrainMapFile.FileType.CSV
    "CSV"
    EXCEL = add_strain_map_request.StrainMapFile.FileType.Excel
    "Excel"
    IMAGE = add_strain_map_request.StrainMapFile.FileType.Image
    "Image"


class StrainMapLegendOrientation:
    """Constants for legend orientation in the add strain maps request."""

    __legend_orientation = add_strain_map_request.StrainMapFile.StrainMapImageFile.LegendOrientation
    HORIZONTAL = __legend_orientation.Horizontal
    "Horizontal"
    VERTICAL = __legend_orientation.Vertical
    "Vertical"


class ImportGDSIIRequest(BaseModel):
    """Contains the information to import a GDSII project file and any optional config files."""

    gdsii_file: str
    """Full path to the GDSII file (.gds, .sf, or .strm) to be imported."""

    technology_file: Optional[str] = None
    """Full path to the optional technology file (.xml, .tech, or .layermap) to be imported."""

    layer_map_file: Optional[str] = None
    """Full path to the optional layer map file (.map) to be imported."""

    project: Optional[str] = None
    """Sherlock project name. If empty, the filename will be used for the project name."""

    cca_name: Optional[str] = None
    """Project CCA name. If empty, the filename will be used for the CCA name."""

    guess_part_properties: Optional[bool] = False
    """Option to guess part properties."""

    polyline_simplification_enabled: Optional[bool] = False
    """Option to enable polyline simplification."""

    polyline_tolerance: Optional[float] = 0.0
    """Polyline simplification tolerance, if enabled."""

    polyline_tolerance_units: Optional[str] = None
    """Polyline simplification tolerance units, if enabled."""

    @field_validator("gdsii_file", "technology_file", "layer_map_file")
    @classmethod
    def str_validation(cls, value: Optional[str], info):
        """Validate string fields listed."""
        if value is None or value.strip() == "":
            raise ValueError(f"{info.field_name} is invalid because it is None or empty.")
        return basic_str_validator(value, info.field_name)

    @field_validator("project", "cca_name")
    @classmethod
    def optional_str_validation(cls, value: Optional[str], info):
        """Allow empty strings for project and cca_name."""
        return optional_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> project_service.ImportGDSIIRequest:
        return project_service.ImportGDSIIRequest(
            gdsiiFile=self.gdsii_file,
            technologyFile=self.technology_file or "",
            layerMapFile=self.layer_map_file or "",
            project=self.project or "",
            ccaName=self.cca_name or "",
            guessPartProperties=self.guess_part_properties,
            polylineSimplificationEnabled=self.polyline_simplification_enabled,
            polylineTolerance=self.polyline_tolerance,
            polylineToleranceUnits=self.polyline_tolerance_units or "",
        )


class OutlineFileType(Enum):
    """Constants for File Type in the Update Outline request."""

    CSV_EXCEL = project_service.OutlineFile.FileType.CsvExcel
    "CSV Excel"
    GERBER = project_service.OutlineFile.FileType.Gerber
    "Gerber"
    IPC2581 = project_service.OutlineFile.FileType.IPC2581
    "IPC2581"


class CsvExcelOutlineFile(BaseModel):
    """Contains the properties for a CSV or Excel outline file."""

    header_row_count: int
    """Number of rows before the column header in the file"""
    x_location_column: str
    """X location column name"""
    y_location_column: str
    """Y location column name"""
    location_units: str
    """Units for location values"""

    def _convert_to_grpc(self) -> project_service.OutlineFile.CsvExcelOutlineFile:
        return project_service.OutlineFile.CsvExcelOutlineFile(
            headerRowCount=self.header_row_count,
            xLocationColumn=self.x_location_column,
            yLocationColumn=self.y_location_column,
            locationUnits=self.location_units,
        )

    @field_validator("header_row_count")
    @classmethod
    def non_negative_int_validation(cls, value: int, info: ValidationInfo):
        """Validate integer fields listed contain non-negative values."""
        if value < 0:
            raise ValueError(f"{info.field_name} must be greater than or equal to 0.")
        return value

    @field_validator("x_location_column", "y_location_column", "location_units")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class GerberOutlineFile(BaseModel):
    """Contains the properties for a Gerber outline file."""

    parse_decimal_first: bool = False
    """Indicates to parse decimals based on the length of the format specifier."""

    def _convert_to_grpc(self) -> project_service.OutlineFile.GerberOutlineFile:
        return project_service.OutlineFile.GerberOutlineFile(
            parseDecimalFirst=self.parse_decimal_first,
        )


class OutlineFile(BaseModel):
    """Contains the properties for an outline file."""

    file_name: str
    """The filename of the outline file."""
    file_comment: Optional[str] = None
    """The comment for the outline file."""
    file_type: OutlineFileType
    """The type of outline file."""
    outline_file_data: Optional[Union[CsvExcelOutlineFile, GerberOutlineFile]] = None
    """Specific outline file type properties"""
    cca_names: list[str]
    """The list of CCA names for the outline file."""

    def _convert_to_grpc(self) -> project_service.OutlineFile:
        grpc_outline_file = project_service.OutlineFile()

        grpc_outline_file.fileName = self.file_name

        if self.file_comment is not None:
            grpc_outline_file.fileComment = self.file_comment

        grpc_outline_file.fileType = self.file_type.value

        grpc_outline_file.cca.extend(self.cca_names)

        if self.file_type == OutlineFileType.CSV_EXCEL:
            if self.outline_file_data is None or not isinstance(
                self.outline_file_data, CsvExcelOutlineFile
            ):
                raise ValueError(
                    "CsvExcel file outline file data is required for CSV Excel outline files."
                )

            grpc_outline_file.csvExcelFile.CopyFrom(self.outline_file_data._convert_to_grpc())

        elif self.file_type == OutlineFileType.GERBER and (
            self.outline_file_data is None
            or not isinstance(self.outline_file_data, GerberOutlineFile)
        ):
            raise ValueError("Gerber file outline file data is required for Gerber outline files.")

            grpc_outline_file.gerberFile.CopyFrom(self.outline_file_data._convert_to_grpc())

        return grpc_outline_file

    @field_validator("file_name")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    @field_validator("file_comment")
    @classmethod
    def optional_str_validation(cls, value: Optional[str], info):
        """Allow empty strings for file_comment."""
        return optional_str_validator(value, info.field_name)

    @field_validator("cca_names")
    @classmethod
    def validate_non_empty_cca_names(cls, value: list[str], info: ValidationInfo):
        """Validate cca_names contains at least one name and none are empty."""
        return basic_list_str_validator(value, info.field_name)


class AddOutlineFileRequest(BaseModel):
    """Contains the information to add outline files to a project."""

    project: str
    """Sherlock project name."""
    outline_files: list[OutlineFile]
    """List of outline files to add."""

    def _convert_to_grpc(self) -> project_service.AddOutlineFileRequest:
        if not self.outline_files:
            raise ValueError("At least one outline file is required.")

        request = project_service.AddOutlineFileRequest()
        request.project = self.project
        for outline_file in self.outline_files:
            request.outlineFiles.append(outline_file._convert_to_grpc())
        return request

    @field_validator("project")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class CopperGerberFile(BaseModel):
    """Properties specific to a Gerber copper file."""

    parse_decimal_first_enabled: Optional[bool] = False
    """Whether to parse decimal values before other formats in the Gerber file."""


class CopperImageFile(BaseModel):
    """Properties specific to an image-based copper file."""

    image_type: Optional[project_service.CopperFile.ImageType.ValueType] = (
        project_service.CopperFile.ImageType.Background
    )
    """Indicates whether the image represents a background or foreground layer."""

    image_color: Optional[str] = ""
    """Optional color description or label for the image (e.g., 'black', 'white')."""

    """Allow non-standard types like Protobuf enums in Pydantic models."""
    model_config = {"arbitrary_types_allowed": True}


class CopperFile(BaseModel):
    """Metadata and options for a copper file to be imported into a project."""

    file_name: str
    """The name of the file being imported."""

    file_type: Optional[project_service.CopperFile.FileType.ValueType] = (
        project_service.CopperFile.FileType.EDB
    )
    """The format/type of the copper file (e.g., Gerber, ODB++, IPC2581)."""

    file_comment: Optional[str] = ""
    """Optional comment or description for the copper file."""

    copper_layer: str
    """The name of the copper layer this file is associated with."""

    polarity: Optional[project_service.CopperFile.Polarity.ValueType] = (
        project_service.CopperFile.Polarity.Positive
    )
    """Indicates whether the copper file uses positive or negative polarity."""

    layer_snapshot_enabled: Optional[bool] = False
    """Enable or disable the generation of a layer snapshot for this copper file."""

    cca: List[str] = []
    """List of CCA (circuit card assembly) names associated with this copper file."""

    gerber_file: Optional[CopperGerberFile] = None
    """Optional settings specific to Gerber file import."""

    image_file: Optional[CopperImageFile] = None
    """Optional settings specific to image-based file import."""

    """Allow non-standard types like Protobuf enums in Pydantic models."""
    model_config = {"arbitrary_types_allowed": True}

    def _convert_to_grpc(self) -> project_service.CopperFile:
        copper_file_msg = project_service.CopperFile()
        copper_file_msg.fileName = self.file_name
        copper_file_msg.fileType = self.file_type
        copper_file_msg.fileComment = self.file_comment or ""
        copper_file_msg.copperLayer = self.copper_layer
        copper_file_msg.polarity = self.polarity
        copper_file_msg.layerSnapshotEnabled = self.layer_snapshot_enabled
        copper_file_msg.cca.extend(self.cca)

        if self.gerber_file:
            copper_file_msg.gerberFile.parseDecimalFirstEnabled = (
                self.gerber_file.parse_decimal_first_enabled
            )

        if self.image_file:
            copper_file_msg.imageFile.imageType = self.image_file.image_type
            copper_file_msg.imageFile.imageColor = self.image_file.image_color or ""

        return copper_file_msg


class ImportCopperFile(BaseModel):
    """Representation of a copper file and its associated metadata for import."""

    copper_file: str
    """The path to the copper file being imported."""

    copper_file_properties: CopperFile
    """Metadata and configuration options for this copper file."""

    @field_validator("copper_file")
    @classmethod
    def copper_file_validator(cls, value: str, info):
        """Validate that the file path is not empty."""
        if value.strip() == "":
            raise ValueError(f"{info.field_name} cannot be empty.")
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> project_service.ImportCopperFilesRequest.ImportCopperFile:
        grpc_import_copper_file = project_service.ImportCopperFilesRequest.ImportCopperFile()
        grpc_import_copper_file.copperFile = self.copper_file
        grpc_import_copper_file.copperFileProperties.CopyFrom(
            self.copper_file_properties._convert_to_grpc()
        )
        return grpc_import_copper_file


class ImportCopperFilesRequest(BaseModel):
    """Request to import multiple copper files into a given project."""

    project: str
    """The name of the project into which the copper files will be imported."""

    copper_files: List[ImportCopperFile]
    """List of copper files with metadata to import into the project."""

    @field_validator("project")
    @classmethod
    def project_validator(cls, value: str, info):
        """Validate that the project name is not empty."""
        if value.strip() == "":
            raise ValueError(f"{info.field_name} cannot be empty.")
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(self) -> project_service.ImportCopperFilesRequest:
        request = project_service.ImportCopperFilesRequest()
        request.project = self.project
        for copper_file in self.copper_files:
            request.copperFiles.append(copper_file._convert_to_grpc())
        return request



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# © 2023 - 2024 ANSYS, Inc. All rights reserved
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""PySherlock logger."""
from datetime import datetime
import logging
from logging.handlers import TimedRotatingFileHandler
import sys

LOG_LEVEL = logging.DEBUG
FILE_NAME = "PySherlock.log"

# Formatting
STDOUT_MSG_FORMAT = logging.Formatter("%(levelname)s - %(module)s - %(funcName)s - %(message)s")
FILE_MSG_FORMAT = STDOUT_MSG_FORMAT

DEFAULT_STDOUT_HEADER = """
LEVEL - INSTANCE NAME - MODULE - FUNCTION - MESSAGE
"""
DEFAULT_FILE_HEADER = DEFAULT_STDOUT_HEADER

NEW_SESSION_HEADER = f"""
===============================================================================
       NEW SESSION - {datetime.now().strftime("%m/%d/%Y, %H:%M:%S")}
==============================================================================="""


def _get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(STDOUT_MSG_FORMAT)
    return console_handler


def _get_file_handler():
    file_handler = TimedRotatingFileHandler(FILE_NAME, when="midnight")
    file_handler.setFormatter(STDOUT_MSG_FORMAT)
    file_handler.stream.write(NEW_SESSION_HEADER)
    file_handler.stream.write(DEFAULT_FILE_HEADER)
    return file_handler


class Logger:
    """Provides the PySherlock logger."""

    def __init__(self, logger_name: str):
        """Initialize logger."""
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(LOG_LEVEL)
        self.logger.addHandler(_get_console_handler())
        self.logger.addHandler(_get_file_handler())
        self.debug = self.logger.debug
        self.info = self.logger.info
        self.warning = self.logger.warning
        self.error = self.logger.error
        self.critical = self.logger.critical
        self.log = self.logger.log



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing all analysis capabilities."""
from typing import Optional

import grpc

from ansys.sherlock.core.types.analysis_types import (
    ElementOrder,
    ModelSource,
    RunAnalysisRequestAnalysisType,
    RunStrainMapAnalysisRequestAnalysisType,
    UpdateComponentFailureMechanismPropsRequest,
    UpdatePcbModelingPropsRequestAnalysisType,
    UpdatePcbModelingPropsRequestPcbMaterialModel,
    UpdatePcbModelingPropsRequestPcbModelType,
    UpdatePTHFatiguePropsRequest,
    UpdateSemiconductorWearoutAnalysisPropsRequest,
)

try:
    import SherlockAnalysisService_pb2
    import SherlockAnalysisService_pb2_grpc
    import SherlockCommonService_pb2
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockCommonService_pb2
    from ansys.api.sherlock.v0 import SherlockAnalysisService_pb2
    from ansys.api.sherlock.v0 import SherlockAnalysisService_pb2_grpc

from ansys.sherlock.core import LOG
from ansys.sherlock.core.errors import (
    SherlockGetPartsListValidationAnalysisPropsError,
    SherlockNoGrpcConnectionException,
    SherlockRunAnalysisError,
    SherlockRunStrainMapAnalysisError,
    SherlockUpdateHarmonicVibePropsError,
    SherlockUpdateICTAnalysisPropsError,
    SherlockUpdateMechanicalShockPropsError,
    SherlockUpdateNaturalFrequencyPropsError,
    SherlockUpdatePartListValidationAnalysisPropsError,
    SherlockUpdatePartModelingPropsError,
    SherlockUpdatePcbModelingPropsError,
    SherlockUpdateRandomVibePropsError,
    SherlockUpdateSolderFatiguePropsError,
)
from ansys.sherlock.core.grpc_stub import GrpcStub
from ansys.sherlock.core.utils.version_check import require_version


class Analysis(GrpcStub):
    """Contains all analysis capabilities."""

    def __init__(self, channel: grpc.Channel, server_version: int):
        """Initialize a gRPC stub for the Sherlock Analysis service."""
        super().__init__(channel, server_version)
        self.stub = SherlockAnalysisService_pb2_grpc.SherlockAnalysisServiceStub(channel)
        self.FIELD_NAMES = {
            "analysisTemp": "analysis_temp",
            "analysisTemp (optional)": "analysis_temp",
            "analysisTempUnits": "analysis_temp_units",
            "analysisTempUnits (optional)": "analysis_temp_units",
            "criticalStrainShock": "critical_strain_shock",
            "criticalStrainShockUnits": "critical_strain_shock_units",
            "filterByEventFrequency": "filter_by_event_frequency",
            "forceModelRebuild": "force_model_rebuild",
            "harmonicVibeDamping": "harmonic_vibe_damping",
            "harmonicVibeCount": "harmonic_vibe_count",
            "ictApplicationTime": "ict_application_time",
            "ictApplicationTimeUnits": "ict_application_time_units",
            "ictNumberOfEvents": "ict_number_of_events",
            "ictResultCount": "ict_result_count",
            "modelSource": "model_source",
            "naturalFreqCount": "natural_freq_count",
            "naturalFreqMin": "natural_freq_min",
            "naturalFreqMinUnits": "natural_freq_min_units",
            "naturalFreqMax": "natural_freq_max",
            "naturalFreqMaxUnits": "natural_freq_max_units",
            "partTemp": "part_temp",
            "partTempUnits": "part_temp_units",
            "partValidationEnabled": "part_validation_enabled",
            "performNFFreqRangeCheck": "perform_nf_freq_range_check",
            "randomVibeDamping": "random_vibe_damping",
            "requireMaterialAssignmentEnabled": "require_material_assignment_enabled",
            "reuseModalAnalysis": "reuse_modal_analysis",
            "shockResultCount": "shock_result_count",
            "solderMaterial": "solder_material",
            "strainMapNaturalFreqs": "strain_map_natural_freqs",
            "usePartTempRiseMin": "use_part_temp_rise_min",
        }

    def _translate_field_names(self, names_list: list[str]) -> list[str]:
        names = []
        for name in list(names_list):
            names.append(self.FIELD_NAMES.get(name))

        return names

    @staticmethod
    def _add_analyses(
        request: SherlockAnalysisService_pb2.RunAnalysisRequest,
        analyses: list[tuple[RunAnalysisRequestAnalysisType, tuple[str, str]]],
    ):
        """Add analyses."""
        for a in analyses:
            analysis = request.analyses.add()
            analysis.type = a[0]
            for p in a[1]:
                phase = analysis.phases.add()
                phase.name = p[0]
                for e in p[1]:
                    event = phase.events.add()
                    event.name = e

    @require_version()
    def run_analysis(
        self,
        project: str,
        cca_name: str,
        analyses: list[tuple[RunAnalysisRequestAnalysisType, tuple[str, str]]],
    ) -> int:
        """Run one or more Sherlock analyses.

        Available Since: 2021R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        analyses: list of ``elements``

            - elements: list[tuple[RunAnalysisRequestAnalysisType, tuple[str, str]]]
                Tuples (``type``, ``event``)

                - analysis_type: RunAnalysisRequestAnalysisType
                    Type of analysis to run.

                - event: list[tuple[str, str]]
                    Tuples (``phase_name``, ``event_name``)

                    - phase_name: str
                        Name of the life cycle phase.
                    - event_name: str
                        Name of the life cycle event.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>    "ODB++ Tutorial.tgz",
        >>>    True,
        >>>    True,
        >>>    True,
        >>>    True,
        >>>    project="Test",
        >>>    cca_name="Card",
        >>> )
        >>> sherlock.analysis.run_analysis(
        >>>    "Test",
        >>>    "Card",
        >>>    [
        >>>        (RunAnalysisRequestAnalysisType.NATURAL_FREQ,
        >>>        [
        >>>            ("Phase 1", ["Harmonic Event"])
        >>>        ]
        >>>        )
        >>>    ]
        >>> )
        """
        try:
            if project == "":
                raise SherlockRunAnalysisError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockRunAnalysisError(message="CCA name is invalid.")
            if not isinstance(analyses, list):
                raise SherlockRunAnalysisError("Analyses argument is invalid.")
            if len(analyses) == 0:
                raise SherlockRunAnalysisError("One or more analyses are missing.")
        except SherlockRunAnalysisError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockAnalysisService_pb2.RunAnalysisRequest(
            project=project,
            ccaName=cca_name,
        )

        self._add_analyses(request, analyses)

        response = self.stub.runAnalysis(request)

        try:
            if response.value == -1:
                raise SherlockRunAnalysisError(response.message)
            else:
                LOG.info(response.message)
                return response.value
        except SherlockRunAnalysisError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def get_harmonic_vibe_input_fields(
        self, model_source: Optional[ModelSource] = None
    ) -> list[str]:
        """Get harmonic vibe property fields based on the user configuration.

        Available Since: 2024R1

        Parameters
        ----------
        model_source: ModelSource, optional
            Model source to get the harmonic vibe property fields from.
            The default is ``None``.

        Returns
        -------
        list[str]
            Harmonic vibe property fields based on the user configuration.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.get_harmonic_vibe_input_fields(ModelSource.GENERATED)
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        message = SherlockAnalysisService_pb2.GetHarmonicVibeInputFieldsRequest(
            modelSource=model_source
        )
        response = self.stub.getHarmonicVibeInputFields(message)

        fields = self._translate_field_names(response.fieldName)
        return fields

    @require_version()
    def update_harmonic_vibe_props(
        self,
        project: str,
        harmonic_vibe_properties: list[dict[str, bool | int | float | str]],
    ) -> int:
        """Update properties for a harmonic vibe analysis.

        Available Since: 2024R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        harmonic_vibe_properties: list[dict[str, bool | int | float | str]]
            Harmonic vibe properties for a CCA consisting of these properties:

            - cca_name: str
                Name of the CCA.
            - model_source: ModelSource
                Model source. The default is ``None``.
            - harmonic_vibe_count: int
                Number of harmonic vibe result layers to generate. The default is ``None``.
            - harmonic_vibe_damping: str
                One or more modal damping ratios. The default is ``None``.
                Separate multiple float values with commas.
            - part_validation_enabled: bool
                Whether to enable part validation. The default is ``None``.
            - require_material_assignment_enabled: bool
                Whether to require material assignment. The default is ``None``.
            - analysis_temp: float
                Temperature. The default is ``None``.
            - analysis_temp_units: str
                Temperature units. The default is ``None``.
                Options are ``"C"``, ``"F"``, and ``"K"``.
            - force_model_rebuild: str
                How to handle rebuilding of the model. The default is ``None``.
                Options are ``"FORCE"`` and ``"AUTO"``.
            - filter_by_event_frequency: bool
                Indicates if harmonic results outside analysis event range are included.
                This parameter is not used for NX Nastran analysis.
            - natural_freq_min: int
                Minimum frequency. The default is ``None``.
                This parameter is for NX Nastran analysis only.
            - natural_freq_min_units: str
                Minimum frequency units. The default is ``None``.
                Options are ``"HZ"``, ``"KHZ"``, ``"MHZ"``, and ``"GHZ"``.
                This parameter is for NX Nastran analysis only.
            - natural_freq_max: int
                Maximum frequency. The default is ``None``.
                This parameter is for NX Nastran analysis only.
            - natural_freq_max_units: str
                Maximum frequency units. The default is ``None``.
                Options are ``"HZ"``, ``"KHZ"``, ``"MHZ"``, and ``"GHZ"``.
                This parameter is for NX Nastran analysis only.
            - reuse_modal_analysis: bool
                Whether to reuse the natural frequency for modal analysis. The
                default is ``None``. This parameter is for NX Nastran analysis only.
            - strain_map_natural_freq: double
                Natural frequency for strain map analysis.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.update_harmonic_vibe_props(
        >>> "Test",
        >>> [{
        >>>     "cca_name": "Card",
        >>>     "model_source": ModelSource.GENERATED,
        >>>     "harmonic_vibe_count": 2,
        >>>     "harmonic_vibe_damping": "0.01, 0.05",
        >>>     "part_validation_enabled": False,
        >>>     "require_material_assignment_enabled": False,
        >>>     "analysis_temp": 20,
        >>>     "analysis_temp_units": "C",
        >>>     "force_model_rebuild": "AUTO",
        >>>     "filter_by_event_frequency": False,
        >>>     "natural_freq_min": 10,
        >>>     "natural_freq_min_units": "Hz",
        >>>     "natural_freq_max": 1000,
        >>>     "natural_freq_max_units": "KHz",
        >>>     "reuse_modal_analysis": True,
        >>>     "strain_map_natural_freq": 100.13,
        >>> },
        >>> ]
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdateHarmonicVibePropsError(message="Project name is invalid.")

            if not isinstance(harmonic_vibe_properties, list):
                raise SherlockUpdateHarmonicVibePropsError(
                    message="Harmonic vibe properties argument is invalid."
                )

            if len(harmonic_vibe_properties) == 0:
                raise SherlockUpdateHarmonicVibePropsError(
                    message="One or more harmonic vibe properties are required."
                )

            request = SherlockAnalysisService_pb2.UpdateHarmonicVibePropsRequest(project=project)

            self._set_update_harmonic_vibe_props_request_properties(
                request, harmonic_vibe_properties
            )
        except SherlockUpdateHarmonicVibePropsError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        response = self.stub.updateHarmonicVibeProps(request)

        try:
            if response.value == -1:
                raise SherlockUpdateHarmonicVibePropsError(response.message)
            else:
                LOG.info(response.message)
                return response.value
        except SherlockUpdateHarmonicVibePropsError as e:
            LOG.error(str(e))
            raise e

    @staticmethod
    def _set_update_harmonic_vibe_props_request_properties(request, harmonic_vibe_properties):
        for i, harmonic_vibe_props in enumerate(harmonic_vibe_properties):
            if not isinstance(harmonic_vibe_props, dict):
                raise SherlockUpdateHarmonicVibePropsError(
                    f"Harmonic vibe props argument is invalid for harmonic vibe properties {i}."
                )

            if "cca_name" not in harmonic_vibe_props.keys():
                raise SherlockUpdateHarmonicVibePropsError(
                    message=f"CCA name is missing for harmonic vibe properties {i}."
                )

            cca_name = harmonic_vibe_props["cca_name"]
            if cca_name == "":
                raise SherlockUpdateHarmonicVibePropsError(
                    message=f"CCA name is invalid for harmonic vibe properties {i}."
                )

            if "model_source" in harmonic_vibe_props.keys():
                model_source = harmonic_vibe_props["model_source"]
            else:
                model_source = None

            if "harmonic_vibe_count" in harmonic_vibe_props.keys():
                harmonic_vibe_count = harmonic_vibe_props["harmonic_vibe_count"]
            else:
                harmonic_vibe_count = None

            if "harmonic_vibe_damping" in harmonic_vibe_props.keys():
                harmonic_vibe_damping = harmonic_vibe_props["harmonic_vibe_damping"]
                if harmonic_vibe_damping is not None:
                    for value in harmonic_vibe_damping.split(","):
                        try:
                            float(value.strip())
                        except ValueError:
                            raise SherlockUpdateHarmonicVibePropsError(
                                message=f"Harmonic vibe damping value is invalid"
                                f" for harmonic vibe properties {i}: " + value.strip()
                            )
            else:
                harmonic_vibe_damping = None

            if "part_validation_enabled" in harmonic_vibe_props.keys():
                part_validation_enabled = harmonic_vibe_props["part_validation_enabled"]
            else:
                part_validation_enabled = None

            if "require_material_assignment_enabled" in harmonic_vibe_props.keys():
                require_material_assignment_enabled = harmonic_vibe_props[
                    "require_material_assignment_enabled"
                ]
            else:
                require_material_assignment_enabled = None

            if "analysis_temp" in harmonic_vibe_props.keys():
                analysis_temp = harmonic_vibe_props["analysis_temp"]
            else:
                analysis_temp = None

            if "analysis_temp_units" in harmonic_vibe_props.keys():
                analysis_temp_units = harmonic_vibe_props["analysis_temp_units"]
            else:
                analysis_temp_units = None

            if "force_model_rebuild" in harmonic_vibe_props.keys():
                force_model_rebuild = harmonic_vibe_props["force_model_rebuild"]
            else:
                force_model_rebuild = None

            if "filter_by_event_frequency" in harmonic_vibe_props.keys():
                filter_by_event_frequency = harmonic_vibe_props["filter_by_event_frequency"]
            else:
                filter_by_event_frequency = None

            if "natural_freq_min" in harmonic_vibe_props.keys():
                natural_freq_min = harmonic_vibe_props["natural_freq_min"]
            else:
                natural_freq_min = None

            if "natural_freq_min_units" in harmonic_vibe_props.keys():
                natural_freq_min_units = harmonic_vibe_props["natural_freq_min_units"]
            else:
                natural_freq_min_units = None

            if "natural_freq_max" in harmonic_vibe_props.keys():
                natural_freq_max = harmonic_vibe_props["natural_freq_max"]
            else:
                natural_freq_max = None

            if "natural_freq_max_units" in harmonic_vibe_props.keys():
                natural_freq_max_units = harmonic_vibe_props["natural_freq_max_units"]
            else:
                natural_freq_max_units = None

            if "reuse_modal_analysis" in harmonic_vibe_props.keys():
                reuse_modal_analysis = harmonic_vibe_props["reuse_modal_analysis"]
            else:
                reuse_modal_analysis = None

            if "strain_map_natural_freq" in harmonic_vibe_props.keys():
                strain_map_natural_freq = harmonic_vibe_props["strain_map_natural_freq"]
            else:
                strain_map_natural_freq = None

            props_request = request.harmonicVibeProperties.add()
            props_request.ccaName = cca_name

            if model_source is not None:
                props_request.modelSource = model_source

            if harmonic_vibe_count is not None:
                props_request.harmonicVibeCount = harmonic_vibe_count

            if harmonic_vibe_damping is not None:
                props_request.harmonicVibeDamping = harmonic_vibe_damping

            if part_validation_enabled is not None:
                props_request.partValidationEnabled = part_validation_enabled

            if require_material_assignment_enabled is not None:
                props_request.requireMaterialAssignmentEnabled = require_material_assignment_enabled

            if analysis_temp is not None:
                props_request.analysisTemp = analysis_temp

            if analysis_temp_units is not None:
                props_request.analysisTempUnits = analysis_temp_units

            if force_model_rebuild is not None:
                props_request.forceModelRebuild = force_model_rebuild

            if filter_by_event_frequency is not None:
                props_request.filterByEventFrequency = filter_by_event_frequency

            if natural_freq_min is not None:
                props_request.naturalFreqMin = natural_freq_min

            if natural_freq_min_units is not None:
                props_request.naturalFreqMinUnits = natural_freq_min_units

            if natural_freq_max is not None:
                props_request.naturalFreqMax = natural_freq_max

            if natural_freq_max_units is not None:
                props_request.naturalFreqMaxUnits = natural_freq_max_units

            if reuse_modal_analysis is not None:
                props_request.reuseModalAnalysis = reuse_modal_analysis

            if strain_map_natural_freq is not None:
                props_request.strainMapNaturalFreq = strain_map_natural_freq

    @require_version(241)
    def get_ict_analysis_input_fields(self) -> list[str]:
        """Get ICT analysis property fields based on the user configuration.

        Available Since: 2024R1

        Returns
        -------
        list[str]
            ICT analysis property fields based on the user configuration.
            ``None`` if not connected to Sherlock.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.analysis.get_ict_analysis_input_fields()
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        message = SherlockAnalysisService_pb2.GetICTAnalysisInputFieldsRequest()
        response = self.stub.getICTAnalysisInputFields(message)

        fields = self._translate_field_names(response.fieldName)
        LOG.info(fields)

        return fields

    @require_version(241)
    def update_ict_analysis_props(
        self,
        project: str,
        ict_analysis_properties: list[dict[str, bool | float | int | str]],
    ) -> int:
        """Update properties for an ICT analysis.

        Available Since: 2024R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        ict_analysis_properties: list[dict[str, bool | float | int | str]]
            ICT analysis properties for a CCA consisting of these properties:

            - cca_name: str
                Name of the CCA.
            - ict_application_time: float
                Specifies the amount of time to complete one ICT event.
            - ict_application_time_units: str
                Application time units.
                Options are ``"ms"``, ``"sec"``, ``"min"``, ``"hr"``, ``"day"``, ``"year"``.
            - ict_number_of_events: int
                Specifies the number of events to apply to the application time when computing
                the time to failure for a component.
            - part_validation_enabled: bool
                Whether to enable part validation. The default is ``None``.
            - require_material_assignment_enabled: bool
                Whether to require material assignment. The default is ``None``.
            - ict_result_count: int
                The number of ICT result layers to generate. This parameter is for use with
                thermal analysis.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.update_ict_analysis_props(
        >>> "Test",
        >>> [{
        >>>     "cca_name": "Card",
        >>>     "ict_application_time": 2,
        >>>     "ict_application_time_units": "sec",
        >>>     "ict_number_of_events": 10,
        >>>     "part_validation_enabled": False,
        >>>     "require_material_assignment_enabled": False,
        >>> },
        >>> ]
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdateICTAnalysisPropsError(message="Project name is invalid.")

            if not isinstance(ict_analysis_properties, list):
                raise SherlockUpdateICTAnalysisPropsError(
                    message="ICT analysis properties argument is invalid."
                )

            if len(ict_analysis_properties) == 0:
                raise SherlockUpdateICTAnalysisPropsError(
                    message="One or more ICT analysis properties are required."
                )

            request = SherlockAnalysisService_pb2.UpdateICTAnalysisPropsRequest(project=project)

            for i, ict_analysis_props in enumerate(ict_analysis_properties):
                if not isinstance(ict_analysis_props, dict):
                    raise SherlockUpdateICTAnalysisPropsError(
                        f"ICT analysis props argument is invalid for ICT analysis properties {i}."
                    )

                if "cca_name" not in ict_analysis_props.keys():
                    raise SherlockUpdateICTAnalysisPropsError(
                        message=f"CCA name is missing for ICT analysis properties {i}."
                    )

                cca_name = ict_analysis_props["cca_name"]
                if cca_name == "":
                    raise SherlockUpdateICTAnalysisPropsError(
                        message=f"CCA name is invalid for ICT analysis properties {i}."
                    )

                props_request = request.ictAnalysisProperties.add()
                props_request.ccaName = cca_name

                if "ict_analysis_count" in ict_analysis_props.keys():
                    props_request.ictAnalysisCount = ict_analysis_props["ict_analysis_count"]

                if "ict_application_time" in ict_analysis_props.keys():
                    props_request.applicationTime = ict_analysis_props["ict_application_time"]

                if "ict_application_time_units" in ict_analysis_props.keys():
                    props_request.applicationTimeUnits = ict_analysis_props[
                        "ict_application_time_units"
                    ]

                if "ict_number_of_events" in ict_analysis_props.keys():
                    props_request.numberOfEvents = ict_analysis_props["ict_number_of_events"]

                if "part_validation_enabled" in ict_analysis_props.keys():
                    props_request.partValidationEnabled = ict_analysis_props[
                        "part_validation_enabled"
                    ]

                if "require_material_assignment_enabled" in ict_analysis_props.keys():
                    props_request.requireMaterialAssignmentEnabled = ict_analysis_props[
                        "require_material_assignment_enabled"
                    ]

                if "force_model_rebuild" in ict_analysis_props.keys():
                    props_request.forceModelRebuild = ict_analysis_props["force_model_rebuild"]

        except SherlockUpdateICTAnalysisPropsError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        response = self.stub.updateICTAnalysisProps(request)

        try:
            if response.value == -1:
                raise SherlockUpdateICTAnalysisPropsError(response.message)
            else:
                LOG.info(response.message)
                return response.value
        except SherlockUpdateICTAnalysisPropsError as e:
            LOG.error(str(e))
            raise e

    @require_version(241)
    def get_mechanical_shock_input_fields(
        self, model_source: Optional[ModelSource] = None
    ) -> list[str]:
        """Get mechanical shock property fields based on the user configuration.

        Available Since: 2024R1

        Parameters
        ----------
        model_source: ModelSource, optional
            Model source to get the random vibe property fields from.
            Only GENERATED is supported.
            Default is ``None``.

        Returns
        -------
        list[str]
            Mechanical shock property fields based on the user configuration.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.get_mechanical_shock_input_fields(ModelSource.GENERATED)
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        message = SherlockAnalysisService_pb2.GetMechanicalShockInputFieldsRequest(
            modelSource=model_source
        )
        response = self.stub.getMechanicalShockInputFields(message)

        fields = self._translate_field_names(response.fieldName)
        LOG.info(fields)

        return fields

    @require_version(241)
    def update_mechanical_shock_props(
        self,
        project: str,
        mechanical_shock_properties: list[dict[str, bool | float | int | str]],
    ) -> int:
        """Update properties for a mechanical shock analysis.

        Available Since: 2024R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        mechanical_shock_properties: list[dict[str, bool | float | int | str]]
            Mechanical shock properties for a CCA consisting of these properties:

            - cca_name: str
                Name of the CCA.
            - model_source: ModelSource, optional
                Model source. The default is ``None``.
            - shock_result_count : int
                Number of mechanical shock result layers to generate.
            - critical_shock_strain: float
                Critical shock strain. The default is ``None``.
            - critical_shock_strain_units: str
                Critical shock strain units. The default is ``None``.
                Options are ``"strain"``, ``"ε"``, and ``"µε"``.
            - part_validation_enabled: bool
                Whether to enable part validation. The default is ``None``.
            - require_material_assignment_enabled: bool
                Whether to require material assignment. The default is ``None``.
            - force_model_rebuild: str
                How to handle rebuilding of the model. The default is ``None``.
                Options are ``"FORCE"`` and ``"AUTO"``.
            - natural_freq_min: int
                Minimum frequency. The default is ``None``.
            - natural_freq_min_units: str
                Minimum frequency units. The default is ``None``.
                Options are ``"HZ"``, ``"KHZ"``, ``"MHZ"``, and ``"GHZ"``.
            - natural_freq_max: int
                Maximum frequency. The default is ``None``.
            - natural_freq_max_units: str
                Maximum frequency units. The default is ``None``.
                Options are ``"HZ"``, ``"KHZ"``, ``"MHZ"``, and ``"GHZ"``.
            - analysis_temp: float
                Temperature. The default is ``None``.
            - analysis_temp_units: str
                Temperature units. The default is ``None``.
                Options are ``"C"``, ``"F"``, and ``"K"``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.update_mechanical_shock_props(
        >>> "Test",
        >>> [{
        >>>     "cca_name": "Card",
        >>>     "model_source": ModelSource.GENERATED,
        >>>     "shock_result_count": 2,
        >>>     "critical_shock_strain": 10,
        >>>     "critical_shock_strain_units": "strain",
        >>>     "part_validation_enabled": True,
        >>>     "require_material_assignment_enabled": False,
        >>>     "force_model_rebuild": "AUTO",
        >>>     "natural_freq_min": 10,
        >>>     "natural_freq_min_units": "Hz",
        >>>     "natural_freq_max": 100,
        >>>     "natural_freq_max_units": "KHz",
        >>>     "analysis_temp": 20,
        >>>     "analysis_temp_units": "F",
        >>> },
        >>> ]
        >>> )

        """
        try:
            if project == "":
                raise SherlockUpdateMechanicalShockPropsError(message="Project name is invalid.")

            if not isinstance(mechanical_shock_properties, list):
                raise SherlockUpdateMechanicalShockPropsError(
                    message="Mechanical shock properties argument is invalid."
                )

            if len(mechanical_shock_properties) == 0:
                raise SherlockUpdateMechanicalShockPropsError(
                    message="One or more mechanical shock properties are required."
                )

            request = SherlockAnalysisService_pb2.UpdateMechanicalShockPropsRequest(project=project)

            for i, mechanical_shock_props in enumerate(mechanical_shock_properties):
                if not isinstance(mechanical_shock_props, dict):
                    raise SherlockUpdateMechanicalShockPropsError(
                        f"Mechanical shock props argument is "
                        f"invalid for mechanical shock properties {i}."
                    )

                if "cca_name" not in mechanical_shock_props.keys():
                    raise SherlockUpdateMechanicalShockPropsError(
                        message=f"CCA name is missing for mechanical shock properties {i}."
                    )

                cca_name = mechanical_shock_props["cca_name"]
                if cca_name == "":
                    raise SherlockUpdateMechanicalShockPropsError(
                        message=f"CCA name is invalid for mechanical shock properties {i}."
                    )

                model_source = mechanical_shock_props.get("model_source", None)
                shock_result_count = mechanical_shock_props.get("shock_result_count", None)
                critical_shock_strain = mechanical_shock_props.get("critical_shock_strain", None)
                critical_shock_strain_units = mechanical_shock_props.get(
                    "critical_shock_strain_units", None
                )
                part_validation_enabled = mechanical_shock_props.get(
                    "part_validation_enabled", None
                )
                require_material_assignment_enabled = mechanical_shock_props.get(
                    "require_material_assignment_enabled", None
                )
                force_model_rebuild = mechanical_shock_props.get("force_model_rebuild", None)
                natural_freq_min = mechanical_shock_props.get("natural_freq_min", None)
                natural_freq_min_units = mechanical_shock_props.get("natural_freq_min_units", None)
                natural_freq_max = mechanical_shock_props.get("natural_freq_max", None)
                natural_freq_max_units = mechanical_shock_props.get("natural_freq_max_units", None)
                analysis_temp = mechanical_shock_props.get("analysis_temp", None)
                analysis_temp_units = mechanical_shock_props.get("analysis_temp_units", None)

                props_request = request.mechanicalShockProperties.add()
                props_request.ccaName = cca_name
                props_request.modelSource = model_source

                if shock_result_count is not None:
                    props_request.shockResultCount = shock_result_count

                if critical_shock_strain is not None:
                    props_request.criticalShockStrain = critical_shock_strain

                if critical_shock_strain_units is not None:
                    props_request.criticalShockStrainUnits = critical_shock_strain_units

                if part_validation_enabled is not None:
                    props_request.partValidationEnabled = part_validation_enabled

                if require_material_assignment_enabled is not None:
                    props_request.requireMaterialAssignmentEnabled = (
                        require_material_assignment_enabled
                    )

                if force_model_rebuild is not None:
                    props_request.forceModelRebuild = force_model_rebuild

                if natural_freq_min is not None:
                    props_request.naturalFreqMin = natural_freq_min

                if natural_freq_min_units is not None:
                    props_request.naturalFreqMinUnits = natural_freq_min_units

                if natural_freq_max is not None:
                    props_request.naturalFreqMax = natural_freq_max

                if natural_freq_max_units is not None:
                    props_request.naturalFreqMaxUnits = natural_freq_max_units

                if analysis_temp is not None:
                    props_request.analysisTemp = analysis_temp

                if analysis_temp_units is not None:
                    props_request.analysisTempUnits = analysis_temp_units

        except SherlockUpdateMechanicalShockPropsError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        response = self.stub.updateMechanicalShockProps(request)

        try:
            if response.value == -1:
                raise SherlockUpdateMechanicalShockPropsError(response.message)
            else:
                LOG.info(response.message)
                return response.value
        except SherlockUpdateMechanicalShockPropsError as e:
            LOG.error(str(e))
            raise e

    @require_version(241)
    def get_solder_fatigue_input_fields(self) -> list[str]:
        """Get solder fatigue property fields based on the user configuration.

        Available Since: 2024R1

        Returns
        -------
        list[str]
            Solder fatigue property fields based on the user configuration.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.get_solder_fatigue_input_fields()
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        message = SherlockAnalysisService_pb2.GetSolderFatigueInputFieldsRequest()
        response = self.stub.getSolderFatigueInputFields(message)

        fields = self._translate_field_names(response.fieldName)
        return fields

    @require_version(241)
    def update_solder_fatigue_props(
        self,
        project: str,
        solder_fatigue_properties: list[dict[str, bool | float | str]],
    ) -> int:
        """Update properties for a solder fatigue analysis.

        Available Since: 2024R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        solder_fatigue_properties: list[dict[str, bool | float | str]]
            Mechanical shock properties for a CCA consisting of these properties:

            - cca_name: str
                Name of the CCA.
            - solder_material: str
                Solder material. The default is ``None``.
            - part_temp: float
                Part temperature. The default is ``None``.
            - part_temp_units: str
                Part temperature units. The default is ``None``.
            - use_part_temp_rise_min: bool
                whether to apply min temp rise. The default is ``None``.
            - part_validation_enabled: bool
                Whether to enable part validation. The default is ``None``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.update_solder_fatigue_props(
        >>> "Test",
        >>> [{
        >>>     "cca_name": "Card",
        >>>     "solder_material": "TIN-LEAD (63SN37PB)",
        >>>     "part_temp": 70,
        >>>     "part_temp_units": "F",
        >>>     "use_part_temp_rise_min": True,
        >>>     "part_validation_enabled": True
        >>> },
        >>> ]
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdateSolderFatiguePropsError(message="Project name is invalid.")

            if not isinstance(solder_fatigue_properties, list):
                raise SherlockUpdateSolderFatiguePropsError(
                    message="Solder fatigue properties argument is invalid."
                )

            if len(solder_fatigue_properties) == 0:
                raise SherlockUpdateSolderFatiguePropsError(
                    message="One or more solder fatigue properties are required."
                )

            request = SherlockAnalysisService_pb2.UpdateSolderFatiguePropsRequest(project=project)

            for i, solder_fatigue_props in enumerate(solder_fatigue_properties):
                if not isinstance(solder_fatigue_props, dict):
                    raise SherlockUpdateSolderFatiguePropsError(
                        f"Solder fatigue props argument is invalid "
                        f"for solder fatigue properties {i}."
                    )

                if "cca_name" not in solder_fatigue_props.keys():
                    raise SherlockUpdateSolderFatiguePropsError(
                        message=f"CCA name is missing for solder fatigue properties {i}."
                    )

                cca_name = solder_fatigue_props["cca_name"]
                if cca_name == "":
                    raise SherlockUpdateSolderFatiguePropsError(
                        message=f"CCA name is invalid for solder fatigue properties {i}."
                    )

                solder_material = solder_fatigue_props.get("solder_material", None)
                part_temp = solder_fatigue_props.get("part_temp", None)
                part_temp_units = solder_fatigue_props.get("part_temp_units", None)
                use_part_temp_rise_min = solder_fatigue_props.get("use_part_temp_rise_min", None)
                part_validation_enabled = solder_fatigue_props.get("part_validation_enabled", None)

                props_request = request.solderFatigueProperties.add()
                props_request.ccaName = cca_name

                if solder_material is not None:
                    props_request.solderMaterial = solder_material

                if part_temp is not None:
                    props_request.partTemp = part_temp

                if part_temp_units is not None:
                    props_request.partTempUnits = part_temp_units

                if use_part_temp_rise_min is not None:
                    props_request.partTempRiseMinEnabled = use_part_temp_rise_min

                if part_validation_enabled is not None:
                    props_request.partValidationEnabled = part_validation_enabled

        except SherlockUpdateSolderFatiguePropsError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        response = self.stub.updateSolderFatigueProps(request)

        try:
            if response.value == -1:
                raise SherlockUpdateSolderFatiguePropsError(response.message)
            else:
                LOG.info(response.message)
                return response.value
        except SherlockUpdateSolderFatiguePropsError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def get_random_vibe_input_fields(self, model_source: Optional[ModelSource] = None) -> list[str]:
        """Get random vibe property fields based on the user configuration.

        Available Since: 2023R2

        Parameters
        ----------
        model_source: ModelSource, optional
            Model source to get the random vibe property fields from.
            The default is ``None``.

        Returns
        -------
        list[str]
            Random vibe input field property names based on the user configuration.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.get_random_vibe_input_fields(ModelSource.STRAIN_MAP)
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        message = SherlockAnalysisService_pb2.GetRandomVibeInputFieldsRequest(
            modelSource=model_source
        )
        response = self.stub.getRandomVibeInputFields(message)

        fields = self._translate_field_names(response.fieldName)
        return fields

    @require_version()
    def update_random_vibe_props(
        self,
        project: str,
        cca_name: str,
        random_vibe_damping: Optional[str] = None,
        natural_freq_min: Optional[float] = None,
        natural_freq_min_units: Optional[str] = None,
        natural_freq_max: Optional[float] = None,
        natural_freq_max_units: Optional[str] = None,
        analysis_temp: Optional[float] = None,
        analysis_temp_units: Optional[str] = None,
        part_validation_enabled: Optional[bool] = None,
        force_model_rebuild: Optional[str] = None,
        reuse_modal_analysis: Optional[bool] = None,
        perform_nf_freq_range_check: Optional[bool] = None,
        require_material_assignment_enabled: Optional[bool] = None,
        model_source: Optional[ModelSource] = None,
        strain_map_natural_freqs: Optional[str] = None,
    ) -> int:
        """Update properties for a random vibe analysis.

        Available Since: 2024R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        random_vibe_damping: str, optional
            One or more modal damping ratios. The default is ``None``.
            Separate multiple float values with commas.
        natural_freq_min: float, optional
            Minimum frequency. The default is ``None``.
            This parameter is for NX Nastran analysis only.
        natural_freq_min_units: str, optional
            Minimum frequency units. The default is ``None``.
            Options are ``"HZ"``, ``"KHZ"``, ``"MHZ"``, and ``"GHZ"``.
            This parameter is for NX Nastran analysis only.
        natural_freq_max: float, optional
            Maximum frequency. The default is ``None``.
            This parameter is for NX Nastran analysis only.
        natural_freq_max_units: str, optional
            Maximum frequency units. The default is ``None``.
            Options are ``"HZ"``, ``"KHZ"``, ``"MHZ"``, and ``"GHZ"``.
            This parameter is for NX Nastran analysis only.
        analysis_temp: float, optional
            Temperature. The default is ``None``.
        analysis_temp_units: str, optional
            Temperature units. The default is ``None``.
            Options are ``"C"``, ``"F"``, and ``"K"``.
        part_validation_enabled: bool, optional
            Whether to enable part validation. The default is ``None``.
        force_model_rebuild: str, optional
            How to handle rebuilding of the model. The default is ``None``.
            Options are ``"FORCE"`` and ``"AUTO"``.
        reuse_modal_analysis: bool, optional
            Whether to reuse the natural frequency for modal analysis. The
            default is ``None``. This parameter is for NX Nastran analysis only.
        perform_nf_freq_range_check: bool, optional
            Whether to perform a frequency range check. The default is ``None``.
            This parameter is for NX Nastran analysis only.
        require_material_assignment_enabled: bool, optional
            Whether to require material assignment. The default is ``None``.
        model_source: ModelSource, optional
            Model source. The default is ``None``.
            This parameter is required for strain map analysis.
        strain_map_natural_freqs: str, optional
            Comma-separated list of natural frequencies. The default is ``None``.
            This parameter is required for strain map analysis.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.update_random_vibe_props(
        >>>     "Test",
        >>>     "Card",
        >>>     random_vibe_damping="0.01, 0.05",
        >>>     analysis_temp=20,
        >>>     analysis_temp_units="C",
        >>>     model_source=ModelSource.STRAIN_MAP
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdateRandomVibePropsError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdateRandomVibePropsError(message="CCA name is invalid.")
            if random_vibe_damping == "":
                raise SherlockUpdateRandomVibePropsError(
                    message="Random vibe damping value is invalid."
                )
            if model_source == SherlockAnalysisService_pb2.ModelSource.STRAIN_MAP and (
                strain_map_natural_freqs is None or strain_map_natural_freqs == ""
            ):
                raise SherlockUpdateRandomVibePropsError(message="Natural frequencies are invalid.")

        except SherlockUpdateRandomVibePropsError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockAnalysisService_pb2.UpdateRandomVibePropsRequest(
            project=project,
            ccaName=cca_name,
            randomVibeDamping=random_vibe_damping,
            naturalFreqMin=natural_freq_min,
            naturalFreqMinUnits=natural_freq_min_units,
            naturalFreqMax=natural_freq_max,
            naturalFreqMaxUnits=natural_freq_max_units,
            analysisTemp=analysis_temp,
            analysisTempUnits=analysis_temp_units,
            partValidationEnabled=part_validation_enabled,
            forceModelRebuild=force_model_rebuild,
            reuseModalAnalysis=reuse_modal_analysis,
            performNFFreqRangeCheck=perform_nf_freq_range_check,
            requireMaterialAssignmentEnabled=require_material_assignment_enabled,
            modelSource=model_source,
        )

        if model_source == SherlockAnalysisService_pb2.ModelSource.STRAIN_MAP:
            request.strainMapNaturalFreqs = strain_map_natural_freqs

        response = self.stub.updateRandomVibeProps(request)

        try:
            if response.value == -1:
                raise SherlockUpdateRandomVibePropsError(response.message)
            else:
                LOG.info(response.message)
                return response.value
        except SherlockUpdateRandomVibePropsError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def get_natural_frequency_input_fields(self) -> list[str]:
        """Get natural frequency property fields based on the user configuration.

        Available Since: 2023R2

        Returns
        -------
        list[str]
            Matural frequency property fields based on the user configuration.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.get_natural_frequency_input_fields()
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        message = SherlockAnalysisService_pb2.GetNaturalFrequencyInputFieldsRequest()
        response = self.stub.getNaturalFrequencyInputFields(message)

        fields = self._translate_field_names(response.fieldName)
        return fields

    @require_version()
    def update_natural_frequency_props(
        self,
        project: str,
        cca_name: str,
        natural_freq_count: int,
        natural_freq_min: int,
        natural_freq_min_units: str,
        natural_freq_max: int,
        natural_freq_max_units: str,
        part_validation_enabled: bool,
        require_material_assignment_enabled: bool,
        analysis_temp: Optional[float] = None,
        analysis_temp_units: Optional[str] = None,
    ) -> int:
        """Update properties for a natural frequency analysis.

        Available Since:2023R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        natural_freq_count: int
            Natural frequency result count.
        natural_freq_min: int, optional
            Minimum frequency. This parameter is for NX Nastran analysis only.
        natural_freq_min_units: str, optional
            Minimum frequency units. Options are ``"HZ"``, ``"KHZ"``, ``"MHZ"``, and ``"GHZ"``.
            This parameter is for NX Nastran analysis only.
        natural_freq_max: int, optional
            Maximum frequency. This parameter is for NX Nastran analysis only.
        natural_freq_max_units: str, optional
            Maximum frequency units. Options are ``"HZ"``, ``"KHZ"``, ``"MHZ"``, and ``"GHZ"``.
            This parameter is for NX Nastran analysis only.
        part_validation_enabled: bool
            Whether part validation is enabled.
        require_material_assignment_enabled: bool
            Whether to require material assignment.
        analysis_temp: float, optional
            Temperature.
        analysis_temp_units: str, optional
            Temperature units. Options are ``"C"``, ``"F"``, and ``"K"``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card"
        >>> )
        >>> sherlock.analysis.update_natural_frequency_props(
        >>>     "Test",
        >>>     "Card",
        >>>     natural_freq_count=2,
        >>>     natural_freq_min=10,
        >>>     natural_freq_min_units="HZ",
        >>>     natural_freq_max=100,
        >>>     natural_freq_max_units="HZ",
        >>>     part_validation_enabled=True,
        >>>     require_material_assignment_enabled=False,
        >>>     analysis_temp=25,
        >>>     analysis_temp_units="C"
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdateNaturalFrequencyPropsError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdateNaturalFrequencyPropsError(message="CCA name is invalid.")
        except SherlockUpdateNaturalFrequencyPropsError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockAnalysisService_pb2.UpdateNaturalFrequencyPropsRequest(
            project=project,
            ccaName=cca_name,
            naturalFreqCount=natural_freq_count,
            naturalFreqMin=natural_freq_min,
            naturalFreqMinUnits=natural_freq_min_units,
            naturalFreqMax=natural_freq_max,
            naturalFreqMaxUnits=natural_freq_max_units,
            partValidationEnabled=part_validation_enabled,
            requireMaterialAssignmentEnabled=require_material_assignment_enabled,
            analysisTemp=analysis_temp,
            analysisTempUnits=analysis_temp_units,
        )

        response = self.stub.updateNaturalFrequencyProps(request)

        try:
            if response.value == -1:
                raise SherlockUpdateNaturalFrequencyPropsError(response.message)
            else:
                LOG.info(response.message)
                return response.value
        except SherlockUpdateNaturalFrequencyPropsError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def run_strain_map_analysis(
        self,
        project: str,
        cca_name: str,
        strain_map_analyses: list[list[RunStrainMapAnalysisRequestAnalysisType | list[list[str]]]],
    ) -> int:
        """Run one or more strain map analyses.

        Available Since: 2023R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the main CCA for the analysis.
        strain_map_analyses: list[list[RunStrainMapAnalysisRequestAnalysisType | list[list[str]]]]
            Analyses consisting of these properties:

            - analysis_type: RunStrainMapAnalysisRequestAnalysisType
                Type of analysis to run.
            - event_strain_maps: list
                Strain maps assigned to the desired life cycle events for
                a given PCB side. The list consists of these properties:

              - phase_name: str
                  Life cycle phase name for the strain map assignment.
              - event_name: str
                  Life cycle event name for the strain map assignment.
              - pcb_side: str
                  PCB side for the strain map. Options are ``"TOP"`` and ``"BOTTOM"``.
              - strain_map: str
                  Name of the strain map assigned to the life cycle event.
              - sub_assembly_name: str, optional
                  Name of the subassembly CCA to assign the strain map to.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.types.analysis_types import (
        >>>     RunStrainMapAnalysisRequestAnalysisType
        >>> )
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> analysis_request = SherlockAnalysisService_pb2.RunStrainMapAnalysisRequest
        >>> sherlock.analysis.run_strain_map_analysis(
        >>>     "AssemblyTutorial",
        >>>     "Main Board",
        >>>     [[
        >>>         RunStrainMapAnalysisRequestAnalysisType.RANDOM_VIBE,
        >>>         [["Phase 1", "Random Vibe", "TOP", "MainBoardStrain - Top"],
        >>>          ["Phase 1", "Random Vibe", "BOTTOM", "MainBoardStrain - Bottom"],
        >>>          ["Phase 1", "Random Vibe", "TOP", "MemoryCard1Strain", "Memory Card 1"]],
        >>>     ]]
        >>> )
        """
        try:
            if project == "":
                raise SherlockRunStrainMapAnalysisError(message="Project name is invalid.")

            if cca_name == "":
                raise SherlockRunStrainMapAnalysisError(message="CCA name is invalid.")

            if not isinstance(strain_map_analyses, list):
                raise SherlockRunStrainMapAnalysisError("Analyses argument is invalid.")

            if len(strain_map_analyses) == 0:
                raise SherlockRunStrainMapAnalysisError("One or more analyses are missing.")

            request = SherlockAnalysisService_pb2.RunStrainMapAnalysisRequest(
                project=project,
                ccaName=cca_name,
            )

            for i, analysis in enumerate(strain_map_analyses):
                if not isinstance(analysis, list):
                    raise SherlockRunStrainMapAnalysisError(
                        f"Analyses argument is invalid for strain map analysis {i}."
                    )

                if len(analysis) != 2:
                    raise SherlockRunStrainMapAnalysisError(
                        f"Number of elements ({str(len(analysis))}) is wrong for "
                        f"strain map analysis {i}."
                    )

                analysis_type = analysis[0]
                if analysis_type == "":
                    raise SherlockRunStrainMapAnalysisError(
                        f"Analysis type is missing for strain map analysis {i}."
                    )

                strain_map_analysis_request = request.strainMapAnalyses.add()
                strain_map_analysis_request.type = analysis_type

                if len(analysis[1]) == 0:
                    raise SherlockRunStrainMapAnalysisError(
                        f"One or more event strain maps are missing for strain map analysis {i}."
                    )

                for j, event_strain_map in enumerate(analysis[1]):
                    if not isinstance(event_strain_map, list):
                        raise SherlockRunStrainMapAnalysisError(
                            f"Event strain maps argument is invalid for strain map analysis {i}."
                        )
                    elif len(event_strain_map) < 4:
                        raise SherlockRunStrainMapAnalysisError(
                            f"Number of elements ({str(len(event_strain_map))}) is wrong for "
                            f"event strain map {j} for strain map analysis {i}."
                        )
                    elif event_strain_map[0] == "":
                        raise SherlockRunStrainMapAnalysisError(
                            f"Life phase is missing for event strain map {j} for strain "
                            f"map analysis {i}."
                        )
                    elif event_strain_map[1] == "":
                        raise SherlockRunStrainMapAnalysisError(
                            f"Event name is missing for event strain map {j} for strain "
                            f"map analysis {i}."
                        )
                    elif event_strain_map[2] == "":
                        raise SherlockRunStrainMapAnalysisError(
                            f"PCB side is missing for event strain map {j} for strain "
                            f"map analysis {i}."
                        )
                    elif event_strain_map[3] == "":
                        raise SherlockRunStrainMapAnalysisError(
                            f"Strain map name is missing for event strain map {j} for strain "
                            f"map analysis {i}."
                        )

                    event_strain_map_request = strain_map_analysis_request.eventStrainMaps.add()
                    event_strain_map_request.phaseName = event_strain_map[0]
                    event_strain_map_request.eventName = event_strain_map[1]
                    event_strain_map_request.pcbSide = event_strain_map[2]
                    event_strain_map_request.strainMap = event_strain_map[3]

                    if len(event_strain_map) == 5:
                        event_strain_map_request.subAssemblyName = event_strain_map[4]
        except SherlockRunStrainMapAnalysisError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        response = self.stub.runStrainMapAnalysis(request)

        try:
            if response.value == -1:
                raise SherlockRunStrainMapAnalysisError(response.message)
            else:
                LOG.info(response.message)
                return response.value
        except SherlockRunStrainMapAnalysisError as e:
            LOG.error(str(e))
            raise e

    @require_version()
    def update_pcb_modeling_props(
        self,
        project: str,
        cca_names: list[str],
        analyses: list[
            tuple[
                bool
                | float
                | str
                | UpdatePcbModelingPropsRequestAnalysisType
                | UpdatePcbModelingPropsRequestPcbModelType
                | UpdatePcbModelingPropsRequestPcbMaterialModel
                | ElementOrder,
                ...,
            ]
        ],
    ) -> int:
        """Update FEA PCB Modeling properties for one or more CCAs.

        Available Since: 2023R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_names: list
            Names of the CCAs to be used for the analysis.
        analyses: list[tuple[bool | float | str | UpdatePcbModelingPropsRequestAnalysisType\
                | UpdatePcbModelingPropsRequestPcbModelType\
                | UpdatePcbModelingPropsRequestPcbMaterialModel\
                | ElementOrder, ...]]
            Elements consisting of the following properties:

            - analysis_type: UpdatePcbModelingPropsRequestAnalysisType
                Type of analysis applied.
            - pcb_model_type: UpdatePcbModelingPropsRequestPcbModelType
                The PCB modeling mesh type.
            - modeling_region_enabled: bool
                Indicates if modeling regions are enabled.
            - pcb_material_model: UpdatePcbModelingPropsRequestPcbMaterialModel
                The PCB modeling PCB model type.
            - pcb_max_materials: Optional[int]
                The number of PCB materials for Uniform Elements and Layered Elements PCB model
                types. Not applicable if PCB model is Uniform or Layered.
            - pcb_elem_order: ElementOrder
                The element order for PCB elements.
            - pcb_max_edge_length: float
                The maximum mesh size for PCB elements.
            - pcb_max_edge_length_units: str
                The length units for the maximum mesh size.
            - pcb_max_vertical: float
                The maximum vertical mesh size for PCB elements.
            - pcb_max_vertical_units: str
                The length units for the maximum vertical mesh size.
            - quads_preferred: bool
                Indicates that the meshing engine should attempt to generate quad-shaped elements
                when creating the mesh.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> update_request = SherlockAnalysisService_pb2.UpdatePcbModelingPropsRequest
        >>> sherlock.analysis.update_pcb_modeling_props(
        >>> "Tutorial Project",
        >>> ["Main Board"],
        >>> [
        >>>     (
        >>>         UpdatePcbModelingPropsRequestAnalysisType.HARMONIC_VIBE,
        >>>         UpdatePcbModelingPropsRequestPcbModelType.BONDED,
        >>>         True,
        >>>         UpdatePcbModelingPropsRequestPcbMaterialModel.UNIFORM,
        >>>         ElementOrder.SOLID_SHELL,
        >>>         6,
        >>>         "mm",
        >>>         3,
        >>>         "mm",
        >>>         True,
        >>>     )
        >>> ]
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdatePcbModelingPropsError(message="Project name is invalid.")
            if not cca_names:
                raise SherlockUpdatePcbModelingPropsError(message="CCA names are invalid.")
            if not analyses:
                raise SherlockUpdatePcbModelingPropsError(message="Analysis input(s) are invalid.")
        except SherlockUpdatePcbModelingPropsError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockAnalysisService_pb2.UpdatePcbModelingPropsRequest(
            project=project,
            ccaNames=cca_names,
        )

        """Add PCB Modeling Props to Request"""
        _analysis = SherlockAnalysisService_pb2.UpdatePcbModelingPropsRequest.Analysis
        uniform = _analysis.PcbMaterialModel.Uniform
        layered = _analysis.PcbMaterialModel.Layered

        for a in analyses:
            analysis = request.analyses.add()
            analysis.type = a[0]
            analysis.modelType = a[1]
            analysis.modelingRegionEnabled = a[2]
            analysis.pcbMaterialModel = a[3]
            if analysis.pcbMaterialModel == uniform or analysis.pcbMaterialModel == layered:
                analysis.pcbElemOrder = a[4]
                analysis.pcbMaxEdgeLength = a[5]
                analysis.pcbMaxEdgeLengthUnits = a[6]
                analysis.pcbMaxVertical = a[7]
                analysis.pcbMaxVerticalUnits = a[8]
                analysis.quadsPreferred = a[9]
            else:
                analysis.pcbMaxMaterials = a[4]
                analysis.pcbElemOrder = a[5]
                analysis.pcbMaxEdgeLength = a[6]
                analysis.pcbMaxEdgeLengthUnits = a[7]
                analysis.pcbMaxVertical = a[8]
                analysis.pcbMaxVerticalUnits = a[9]
                analysis.quadsPreferred = a[10]

        response = self.stub.updatePcbModelingProps(request)

        try:
            if response.value == -1:
                raise SherlockUpdatePcbModelingPropsError(response.message)
            return response.value
        except SherlockUpdatePcbModelingPropsError as e:
            LOG.error(str(e))
            raise e

    @require_version(241)
    def update_part_modeling_props(
        self, project: str, part_modeling_props: dict[str, bool | float | str]
    ) -> int:
        """Update part modeling properties for a given project's CCA.

        Available Since: 2024R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        part_modeling_props: dict[str, bool | float | str]
            Part modeling properties for a CCA consisting of these properties:

            - cca_name: str
                Name of the CCA.
            - part_enabled: bool
                Whether to enable part modeling. All other fields are ignored if disabled.
            - part_min_size: float, optional
                Minimum part size.
            - part_min_size_units: str, optional
                Minimum part size units.
            - part_elem_order: str, optional
                Part element order.
                Options are ``"First Order (Linear)"``, ``"Second Order (Quadratic)"``,
                or ``"Solid Shell"``.
            - part_max_edge_length: float, optional
                Part max edge length.
            - part_max_edge_length_units: str, optional
                Part max edge length units.
            - part_max_vertical: float, optional
                Part max vertical.
            - part_max_vertical_units: str, optional
                Part max vertical units.
            - part_results_filtered: bool, optional
                Whether to enable filtered part results.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.update_part_modeling_props(
        >>> "Test",
        >>> {
        >>>     "cca_name": "Card",
        >>>     "part_enabled": True,
        >>>     "part_min_size": 1,
        >>>     "part_min_size_units": "in",
        >>>     "part_elem_order": "First Order (Linear)",
        >>>     "part_max_edge_length": 1,
        >>>     "part_max_edge_length_units": "in",
        >>>     "part_max_vertical": 1,
        >>>     "part_max_vertical_units": "in",
        >>>     "part_results_filtered": True
        >>> }
        >>> )

        """
        try:
            if project == "":
                raise SherlockUpdatePartModelingPropsError(message="Project name is invalid.")

            if not isinstance(part_modeling_props, dict):
                raise SherlockUpdatePartModelingPropsError(
                    message="Part modeling props argument is invalid."
                )

            if "cca_name" not in part_modeling_props.keys():
                raise SherlockUpdatePartModelingPropsError(message="CCA name is missing.")
            if "part_enabled" not in part_modeling_props.keys():
                raise SherlockUpdatePartModelingPropsError(message="Part enabled is missing.")

            request = SherlockAnalysisService_pb2.UpdatePartModelingRequest(project=project)
            request.ccaName = part_modeling_props["cca_name"]

            if request.ccaName == "":
                raise SherlockUpdatePartModelingPropsError(message="CCA name is invalid.")

            request.partEnabled = part_modeling_props["part_enabled"]

            if request.partEnabled:
                if "part_min_size" in part_modeling_props.keys():
                    request.partMinSize = part_modeling_props["part_min_size"]
                if "part_min_size_units" in part_modeling_props.keys():
                    request.partMinSizeUnits = part_modeling_props["part_min_size_units"]
                if "part_elem_order" in part_modeling_props.keys():
                    request.partElemOrder = part_modeling_props["part_elem_order"]
                if "part_max_edge_length" in part_modeling_props.keys():
                    request.partMaxEdgeLength = part_modeling_props["part_max_edge_length"]
                if "part_max_edge_length_units" in part_modeling_props.keys():
                    request.partMaxEdgeLengthUnits = part_modeling_props[
                        "part_max_edge_length_units"
                    ]
                if "part_max_vertical" in part_modeling_props.keys():
                    request.partMaxVertical = part_modeling_props["part_max_vertical"]
                if "part_max_vertical_units" in part_modeling_props.keys():
                    request.partMaxVerticalUnits = part_modeling_props["part_max_vertical_units"]
                if "part_results_filtered" in part_modeling_props.keys():
                    request.partResultsFiltered = part_modeling_props["part_results_filtered"]

        except SherlockUpdatePartModelingPropsError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        response = self.stub.updatePartModelingProperties(request)

        try:
            if response.value == -1:
                raise SherlockUpdatePartModelingPropsError(response.message)
            else:
                LOG.info(response.message)
                return response.value
        except SherlockUpdatePartModelingPropsError as e:
            LOG.error(str(e))
            raise e

    @require_version(251)
    def update_part_list_validation_analysis_props(
        self,
        project: str,
        properties_per_cca: list[dict[str, bool | str]],
    ) -> int:
        """Update properties for a Part List Validation analysis.

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        properties_per_cca: list[dict[str, bool | str]]
            Part List Validation analysis properties for each CCA consisting of these properties:

            - cca_name: str
                Name of the CCA.
            - process_use_avl: bool
                Whether to use AVL.
            - process_use_wizard: bool
                Whether to use the wizard.
            - process_check_confirmed_properties: bool
                Whether to check confirmed properties.
            - process_check_part_numbers: bool
                Whether to check part numbers.
            - matching_mode: str
                Matching type.
            - avl_require_internal_part_number: bool
                Whether to require an internal part number.
            - avl_require_approved_description: bool
                Whether to require an approved description.
            - avl_require_approved_manufacturer: bool
                Whether to require an approved manufacturer.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.analysis.update_part_list_validation_analysis_props(
        >>> "Test",
        >>> [{
        >>>     "cca_name": "Card",
        >>>     "process_use_avl": True,
        >>>     "process_use_wizard": False,
        >>>     "process_check_confirmed_properties": True,
        >>>     "process_check_part_numbers": True,
        >>>     "matching_mode": "Part",
        >>>     "avl_require_internal_part_number": True,
        >>>     "avl_require_approved_description": False,
        >>>     "avl_require_approved_manufacturer": True,
        >>> },
        >>> ]
        >>> )
        """
        try:
            if not isinstance(properties_per_cca, list):
                raise SherlockUpdatePartListValidationAnalysisPropsError(
                    message="Properties per CCA argument is invalid."
                )

            if len(properties_per_cca) == 0:
                raise SherlockUpdatePartListValidationAnalysisPropsError(
                    message="One or more analysis properties are required."
                )

            request = SherlockAnalysisService_pb2.UpdatePartsListValidationPropsRequest(
                project=project
            )

            for i, analysis_properties in enumerate(properties_per_cca):
                if not isinstance(analysis_properties, dict):
                    raise SherlockUpdatePartListValidationAnalysisPropsError(
                        f"Analysis properties is invalid for index {i}."
                    )

                if "cca_name" not in analysis_properties.keys():
                    raise SherlockUpdatePartListValidationAnalysisPropsError(
                        message=f"CCA name is missing for analysis properties {i}."
                    )

                cca_name = analysis_properties["cca_name"]
                if cca_name == "":
                    raise SherlockUpdatePartListValidationAnalysisPropsError(
                        message=f"CCA name is invalid for analysis properties {i}."
                    )

                request.partsListValidationProperties.add(
                    ccaName=cca_name,
                    processUseAVL=analysis_properties.get("process_use_avl", False),
                    processUseWizard=analysis_properties.get("process_use_wizard", False),
                    processCheckConfirmedProperties=analysis_properties.get(
                        "process_check_confirmed_properties", False
                    ),
                    processCheckPartNumbers=analysis_properties.get(
                        "process_check_part_numbers", False
                    ),
                    matching=analysis_properties.get("matching_mode", "Both"),
                    avlRequireInternalPartNumber=analysis_properties.get(
                        "avl_require_internal_part_number", False
                    ),
                    avlRequireApprovedDescription=analysis_properties.get(
                        "avl_require_approved_description", False
                    ),
                    avlRequireApprovedManufacturer=analysis_properties.get(
                        "avl_require_approved_manufacturer", False
                    ),
                )
        except SherlockUpdatePartListValidationAnalysisPropsError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        response = self.stub.updatePartsListValidationProps(request)

        try:
            if response.value == -1:
                raise SherlockUpdatePartListValidationAnalysisPropsError(response.message)
            else:
                LOG.info(response.message)
                return response.value
        except SherlockUpdatePartListValidationAnalysisPropsError as e:
            LOG.error(str(e))
            raise e

    @require_version(251)
    def get_parts_list_validation_analysis_props(
        self,
        project: str,
        cca_name: str,
    ) -> SherlockAnalysisService_pb2.PartsListValidationPropsResponse:
        """Get properties for a Part List Validation analysis.

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.

        Returns
        -------
        PartsListValidationPropsResponse
            - returnCode: ReturnCode
                - value: int
                    Status code of the response. 0 for success.
                - message: str
                    indicates general errors that occurred while attempting to update parts
            - partLibrary: str
                Part library name
            - processUseAVL: bool
                Process option to use AVL
            - processUseWizard: bool
                Process option to use wizard
            - processCheckConfirmedProperties: bool
                Process option to check confirmed properties
            - processCheckPartNumbers: bool
                Process option to check part numbers
            - matching: MatchingMode
                Matching type
            - avlRequireInternalPartNumber: bool
                AVL option to require internal part number
            - avlRequireApprovedDescription: bool
                AVL option to require approved description
            - avlRequireApprovedManufacturer: bool
                AVL option to require approved manufacturer

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> analysis_props = sherlock.analysis.get_parts_list_validation_analysis_props(
        >>> "Test", "Card"
        >>> )
        """
        try:
            if project == "":
                raise SherlockGetPartsListValidationAnalysisPropsError(
                    message="Project name is invalid."
                )
            if cca_name == "":
                raise SherlockGetPartsListValidationAnalysisPropsError(
                    message="CCA name is invalid."
                )

            request = SherlockAnalysisService_pb2.GetPartsListValidationPropsRequest(
                project=project, ccaName=cca_name
            )

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            response = self.stub.getPartsListValidationProps(request)

            return_code = response.returnCode

            if return_code.value == -1:
                raise SherlockGetPartsListValidationAnalysisPropsError(return_code.message)

            return response
        except SherlockGetPartsListValidationAnalysisPropsError as e:
            LOG.error(str(e))
            raise e

    @require_version(252)
    def update_component_failure_mechanism_analysis_props(
        self,
        request: UpdateComponentFailureMechanismPropsRequest,
    ) -> list[SherlockCommonService_pb2.ReturnCode]:
        r"""Update properties for one or more Component Failure Mechanism analysis.

        Parameters
        ----------
        request: UpdateComponentFailureMechanismPropsRequest
            Contains all the information needed to update the properties for one or more component
            failure mechanism analyses per project.

        Returns
        -------
        list[SherlockCommonService_pb2.ReturnCode]
            Return codes for each request.

        Examples
        --------
        >>> from ansys.sherlock.core.types.analysis_types import (
        >>> ComponentFailureMechanism,
        >>> UpdateComponentFailureMechanismPropsRequest,
        >>> )
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_project_zip_archive(
        >>> project="Assembly Tutorial",
        >>> category="category",
        >>> archive_file=\
        >>>     "C:\\Program Files\\ANSYS Inc\\v252\\sherlock\\tutorial\\Assembly Tutorial.zip",
        >>> )
        >>> update_request1 = ComponentFailureMechanism(
        >>> cca_name="Main Board",
        >>> default_part_temp_rise=1.5,
        >>> default_part_temp_rise_units="K",
        >>> part_temp_rise_min_enabled=True,
        >>> part_validation_enabled=False,
        >>> )
        >>> update_request2 = ComponentFailureMechanism(
        >>> cca_name="Memory Card 1",
        >>> default_part_temp_rise=-3.25,
        >>> default_part_temp_rise_units="F",
        >>> part_temp_rise_min_enabled=False,
        >>> part_validation_enabled=True,
        >>> )
        >>> request = UpdateComponentFailureMechanismPropsRequest(
        >>> project="Test",
        >>> component_failure_mechanism_properties_per_cca=[
        >>>     update_request1,
        >>>     update_request2
        >>> ]
        >>> )
        >>> return_codes = sherlock.analysis.\
        >>>     update_component_failure_mechanism_analysis_props(request)
        >>> for return_code in return_codes:
        >>>     print(f"Return code: value={return_code.value}, message={return_code.message}")
        """
        update_request = request._convert_to_grpc()

        responses = []
        for return_code in self.stub.updateComponentFailureMechanismProps(update_request):
            responses.append(return_code)
        return responses

    @require_version(252)
    def update_semiconductor_wearout_props(
        self,
        request: UpdateSemiconductorWearoutAnalysisPropsRequest,
    ) -> list[SherlockCommonService_pb2.ReturnCode]:
        r"""Update properties for one or more Semiconductor Wearout Analysis.

        Parameters
        ----------
        request: UpdateSemiconductorWearoutAnalysisPropsRequest
            Contains all the information needed to update the properties for one or more
            semiconductor wearout analyses per project.

        Returns
        -------
        list[SherlockCommonService_pb2.ReturnCode]
            Return codes for each request.

        Examples
        --------
        >>> from ansys.sherlock.core.types.analysis_types import (
        >>> SemiconductorWearoutAnalysis,
        >>> UpdateSemiconductorWearoutAnalysisPropsRequest,
        >>> )
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_project_zip_archive(
        >>> project="Assembly Tutorial",
        >>> category="category",
        >>> archive_file=\
        >>>     "C:\\Program Files\\ANSYS Inc\\v252\\sherlock\\tutorial\\Assembly Tutorial.zip",
        >>>)
        >>> update_request1 = SemiconductorWearoutAnalysis(
        >>> cca_name="Main Board",
        >>> max_feature_size=1.5,
        >>> max_feature_size_units="mm",
        >>> part_temp_rise=10.0,
        >>> part_temp_rise_units="C",
        >>> part_temp_rise_min_enabled=True,
        >>> part_validation_enabled=False,
        >>> )
        >>> update_request2 = SemiconductorWearoutAnalysis(
        >>> cca_name="Memory Card 1",
        >>> max_feature_size=2.0,
        >>> max_feature_size_units="mm",
        >>> part_temp_rise=15.0,
        >>> part_temp_rise_units="C",
        >>> part_temp_rise_min_enabled=False,
        >>> part_validation_enabled=True,
        >>> )
        >>> request = UpdateSemiconductorWearoutAnalysisPropsRequest(
        >>> project="Test",
        >>> semiconductor_wearout_analysis_properties=[
        >>>     update_request1,
        >>>     update_request2
        >>> ]
        >>> )
        >>> return_codes = sherlock.analysis.update_semiconductor_wearout_props(request)
        >>> for return_code in return_codes:
        >>>     print(f"Return code: value={return_code.value}, message={return_code.message}")
        """
        update_request = request._convert_to_grpc()

        responses = []
        for return_code in self.stub.updateSemiconductorWearoutAnalysisProps(update_request):
            responses.append(return_code)
        return responses

    @require_version(252)
    def update_PTH_fatigue_props(
        self,
        request: UpdatePTHFatiguePropsRequest,
    ) -> list[SherlockCommonService_pb2.ReturnCode]:
        r"""Update properties for one or more Plated Through Hole (PTH) Fatigue Analyses.

        Parameters
        ----------
        request: UpdatePTHFatiguePropsRequest
            Contains all the information needed to update the properties for one or more
            PTH fatigue analyses in a project.

        Returns
        -------
        list[SherlockCommonService_pb2.ReturnCode]
            Return codes for each update request.

        Examples
        --------
        >>> from ansys.sherlock.core.types.analysis_types import (
        >>> PTHFatiguePropsAnalysis,
        >>> UpdatePTHFatiguePropsRequestAnalysisType,
        >>> UpdatePTHFatiguePropsRequest,
        >>> )
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_project_zip_archive(
        >>> project="Assembly Tutorial",
        >>> category="category",
        >>> archive_file="C:\\Program Files\\ANSYS Inc\\v252\\sherlock\\tutorial\\
        >>>     Assembly Tutorial.zip",
        >>> )
        >>> update_request1 = PTHFatiguePropsAnalysis(
        >>>     cca_name="Main Board",
        >>>     qualification=UpdatePTHFatiguePropsRequestAnalysisType.SUPPLIER,
        >>>     pth_quality_factor="Good",
        >>>     pth_wall_thickness=0.1,
        >>>     pth_wall_thickness_units="mm",
        >>>     min_hole_size=0.5,
        >>>     min_hole_size_units="mm",
        >>>     max_hole_size=1.0,
        >>>     max_hole_size_units="mm",
        >>> )
        >>> update_request2 = PTHFatiguePropsAnalysis(
        >>>     cca_name="Memory Card 1",
        >>>     qualification=UpdatePTHFatiguePropsRequestAnalysisType.PRODUCT,
        >>>     pth_quality_factor="Good",
        >>>     pth_wall_thickness=0.2,
        >>>     pth_wall_thickness_units="mil",
        >>>     min_hole_size=0.7,
        >>>     min_hole_size_units="mil",
        >>>     max_hole_size=1.5,
        >>>     max_hole_size_units="mil",
        >>> )
        >>> request = UpdatePTHFatiguePropsRequest(
        >>> project="Assembly Tutorial",
        >>> pth_fatigue_analysis_properties=[
        >>>     update_request1,
        >>>     update_request2,
        >>> ]
        >>> )
        >>> return_codes = sherlock.analysis.update_PTH_fatigue_props(request)
        >>> for return_code in return_codes:
        >>>     print(f"Return code: value={return_code.value}, message={return_code.message}")
        """
        update_request = request._convert_to_grpc()

        responses = []
        for return_code in self.stub.updatePTHFatigueProps(update_request):
            responses.append(return_code)
        return responses



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for launching Sherlock locally or connecting to a local instance with gRPC."""
import errno
import os
import shlex
import socket
from typing import Optional

import grpc

from ansys.sherlock.core import LOG
from ansys.sherlock.core.common import Common
from ansys.sherlock.core.errors import SherlockCannotUsePortError, SherlockConnectionError
from ansys.sherlock.core.sherlock import Sherlock
from ansys.sherlock.core.utils.cyberchannel import create_channel
from ansys.sherlock.core.utils.version_check import _EARLIEST_SUPPORTED_VERSION

ANSYS_GRPC_CERTIFICATES = "ANSYS_GRPC_CERTIFICATES"
LOCALHOST = "127.0.0.1"
SHERLOCK_DEFAULT_PORT = 9090
SHERLOCK_UDS_SERVICE = "sherlock-grpc"
DEFAULT_CONNECT_TIMEOUT = 120
sherlock_cmd_args = []


def _is_port_available(host: str = LOCALHOST, port: int = SHERLOCK_DEFAULT_PORT):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        try:
            sock.bind((host, port))
            return True
        except socket.error as e:
            if e.errno == errno.EADDRINUSE:
                raise SherlockCannotUsePortError(port, "Port is already in use")

            raise SherlockCannotUsePortError(port, str(e))


def launch(
    host: str = LOCALHOST,
    port: int = SHERLOCK_DEFAULT_PORT,
    single_project_path: str = "",
    sherlock_command_args: str = "",
    year: Optional[int] = None,
    release_number: Optional[int] = None,
    transport_mode: str = "mtls",
    certs_dir: str = None,
    uds_dir: str = None,
    uds_id: str = None,
) -> str:
    r"""Launch Sherlock using the specified host and port for the gRPC connection.

    Available Since: 2025R2

    Parameters
    ----------
    host: str, optional
        IP address to start gRPC on.
        The default is ``"127.0.0.1"``, which is the IP address for the local host.
    port: int, optional
        Port number for the connection.
    single_project_path : str, optional
        Path to the Sherlock project if invoking Sherlock in the single-project mode.
    sherlock_command_args : str, optional
        Additional command arguments for launching Sherlock.
    year: int, optional
        4-digit year of the Sherlock release to launch. If not provided,
        the latest installed version of Sherlock will be launched.
    release_number: int, optional
        Release number of Sherlock to launch. If not provided,
        the latest installed version of Sherlock will be launched.
    transport_mode: str, optional
        See :func:`launch_and_connect` for usage.
    certs_dir: str, optional
        Directory containing the mTLS certificates. Default is "./certs".
    uds_dir: str, optional
        Directory for the UDS socket file. Default is "$HOME/.conn".
    uds_id: str, optional
        Identifier for the UDS socket file name. Default is no identifier.

    Returns
    -------
    str
        Path to the Sherlock installation directory.

    Examples
    --------
    >>> from ansys.sherlock.core import launcher
    >>> project = "C:\\Default Projects Directory\\ODB++ Tutorial"
    >>> ansys_install_path = launcher.launch(
    >>>     port=9092, single_project_path=project, year=2024, release_number=2)
    """
    # Moved import here since it's only used in this function
    import subprocess  # nosec B404

    try:
        _is_port_available(host, port)
    except Exception as e:
        LOG.error(str(e))
        raise e

    try:
        sherlock_launch_cmd, _server_version, ansys_install_path = _get_sherlock_exe_path(
            year=year, release_number=release_number
        )
        args = [sherlock_launch_cmd]

        # Add gRPC options
        if transport_mode in ["insecure", "mtls", "wnua"]:
            args.append(f"-grpcHost={host}")
            args.append(f"-grpcPort={port}")
        args.append(f"--transport-mode={transport_mode}")
        if transport_mode == "mtls":
            if not certs_dir:
                certs_dir = os.getenv(ANSYS_GRPC_CERTIFICATES, "./certs")
            args.append(f"--certs-dir={certs_dir}")
        if transport_mode == "uds":
            if uds_dir:
                args.append(f"--uds-dir={uds_dir}")
            if uds_id:
                args.append(f"--uds-id={uds_id}")

        # Add other options
        if single_project_path:
            args.append("-singleProject")
            args.append(single_project_path)
        if sherlock_command_args:
            args.extend(shlex.split(sherlock_command_args))

        LOG.info(f"Command arguments: {args}")
        # subprocess is used safely here to launch a trusted local executable (Sherlock).
        # Input arguments are internally constructed and shell=False prevents injection.
        subprocess.Popen(args, shell=False)  # nosec B603

        return ansys_install_path
    except Exception as e:
        LOG.error(f"Error launching Sherlock. {e}")
        raise e


def launch_and_connect(
    host: str = LOCALHOST,
    port: int = SHERLOCK_DEFAULT_PORT,
    single_project_path: str = "",
    sherlock_command_args: str = "",
    year: Optional[int] = None,
    release_number: Optional[int] = None,
    timeout: int = DEFAULT_CONNECT_TIMEOUT,
    transport_mode: str = "mtls",
    certs_dir: str = None,
    uds_dir: str = None,
    uds_id: str = None,
) -> tuple[Sherlock, str]:
    """
    Launch Sherlock, start gRPC on a given host and port, and wait until connected to Sherlock.

    Available Since: 2025R2

    Parameters
    ----------
    host: str, optional
        IP address to start gRPC on.
        The default is ``"127.0.0.1"``, which is the IP address for the local host.
    port: int, optional
        Port number for the connection.
    single_project_path : str, optional
        Path to the Sherlock project if invoking Sherlock in the single-project mode.
    sherlock_command_args : str, optional
        Additional command arguments for launching Sherlock.
    year: int, optional
        4-digit year of the Sherlock release to launch. If not provided,
        the latest installed version of Sherlock will be launched.
    release_number: int, optional
        Release number of Sherlock to launch. If not provided,
        the latest installed version of Sherlock will be launched.
    timeout: int, optional
        Maximum time (in seconds) to wait for the connection to Sherlock to be established.
        Default is 120 seconds.
    transport_mode: str, optional
        The gRPC transport mode:
            - "insecure" : unencrypted connection
            - "mtls" : mutual TLS authentication (default)
            - "uds" : Unix Domain Socket
            - "wnua" : Windows Named User Authentication

        See `Securing gRPC connections
        <https://tools.docs.pyansys.com/version/stable/user_guide/secure_grpc.html>`_
        for more details.
    certs_dir: str, optional
        Directory containing the mTLS certificates. Default is "./certs".
    uds_dir: str, optional
        Directory for the UDS socket file. Default is "$HOME/.conn".
    uds_id: str, optional
        Identifier for the UDS socket file name. Default is no identifier.

    Returns
    -------
    tuple[Sherlock, str]
        Sherlock gRPC connection object and the installation directory.

    Examples
    --------
    >>> from ansys.sherlock.core import launcher
    >>> sherlock, ansys_install_path = launcher.launch_and_connect(
    >>>     port=9092, year=2026, release_number=1, transport_mode="wnua")
    """
    ansys_install_path = launch(
        host,
        port,
        single_project_path,
        sherlock_command_args,
        year,
        release_number,
        transport_mode=transport_mode,
        certs_dir=certs_dir,
        uds_dir=uds_dir,
        uds_id=uds_id,
    )
    try:
        sherlock = connect(
            port=port,
            timeout=timeout,
            transport_mode=transport_mode,
            uds_dir=uds_dir,
            uds_id=uds_id,
        )
        return sherlock, ansys_install_path
    except Exception as e:
        LOG.error(f"Error connecting to Sherlock after launch: {e}")
        raise RuntimeError(f"Error connecting to Sherlock after launch: {e}")


def connect(
    port: int = SHERLOCK_DEFAULT_PORT,
    timeout=DEFAULT_CONNECT_TIMEOUT,
    transport_mode: str = "mtls",
    certs_dir: str = None,
    uds_dir: str = None,
    uds_id: str = None,
) -> Sherlock:
    r"""Connect to a local instance of Sherlock.

    Available Since: 2025R2

    Parameters
    ----------
    port: int, optional
        Port number for the connection (ignored if transport_mode is "uds").
        Default is 9090.
    timeout: int, optional
        Maximum time (in seconds) to wait for the connection to Sherlock to be established.
        Default is 120 seconds.
    transport_mode : str, optional
        See :func:`launch_and_connect` for usage.
    certs_dir: str, optional
        Directory containing the mTLS certificates. Default is "./certs".
    uds_dir : str, optional
        Directory for the UDS socket file.
    uds_id : str, optional
        Optional ID for the UDS socket file.

    Returns
    -------
    Sherlock
        The instance of Sherlock.

    Examples
    --------
    >>> from ansys.sherlock.core import launcher
    >>> ansys_install_path = launcher.launch(
    >>>     port=9092, year=2026, release_number=1, transport_mode="wnua")
    >>> sherlock = launcher.connect(port=9092, timeout=60, transport_mode="wnua")
    """
    try:
        channel = _connect_grpc_channel(
            host=LOCALHOST,
            port=port,
            uds_dir=uds_dir,
            uds_id=uds_id,
            transport_mode=transport_mode,
            certs_dir=certs_dir,
        )
        _wait_for_sherlock_grpc_ready(channel, timeout)

        # Create Common without version since the version is unknown
        common = Common(channel=channel, server_version=None)
        server_version = None
        try:
            sherlock_info = common.get_sherlock_info()
            if sherlock_info is not None:
                LOG.info(f"Connected to Sherlock version: {sherlock_info.releaseVersion}")
                server_version = _convert_to_server_version(sherlock_info.releaseVersion)
        except grpc.RpcError as e:
            error_message = e.details() if e.details() else "Unknown error occurred."
            LOG.error(f"Server validation error: {error_message}")
            raise SherlockConnectionError(message=error_message)

        return Sherlock(channel=channel, server_version=server_version)
    except Exception as e:
        LOG.error(f"Error encountered connecting to Sherlock: {str(e)}")
        raise e


def _convert_to_server_version(sherlock_release_version: str) -> int:
    # convert the version returned from Sherlock (e.g. "2025 R1")
    # to the version needed for the API (e.g. 251)
    tokens = sherlock_release_version.split(" ")
    year = _extract_sherlock_version_year(int(tokens[0]))
    minor_version = int(tokens[1][1:])
    server_version = year * 10 + minor_version
    return server_version


def _connect_grpc_channel(
    host: str = LOCALHOST,
    port: int = SHERLOCK_DEFAULT_PORT,
    uds_dir: str = None,
    uds_id: str = None,
    transport_mode: str = "mtls",
    certs_dir: str = None,
) -> grpc.Channel:
    """
    Connect to Sherlock gRPC via UDS, TCP, or other transport modes.

    Parameters
    ----------
    host : str
        Hostname or IP address for TCP connection (ignored for UDS).
    port : int
        Port number for TCP connection.
    uds_dir : str, optional
        Directory for the UDS socket file.
    uds_id : str, optional
        Optional ID for the UDS socket file.
    transport_mode : str
        Transport mode (e.g., "insecure", "wnua", "mtls", "uds").
    certs_dir: str, optional
        Directory containing the mTLS certificates. Default is "./certs".

    Returns
    -------
    grpc.Channel
        A gRPC channel connected to Sherlock.
    """
    try:
        channel = create_channel(
            transport_mode=transport_mode,
            host=host,
            port=port,
            uds_service=SHERLOCK_UDS_SERVICE,
            uds_dir=uds_dir,
            uds_id=uds_id,
            certs_dir=certs_dir,
            grpc_options=[
                ("grpc.max_send_message_length", 100 * 1024 * 1024),
                ("grpc.max_receive_message_length", 100 * 1024 * 1024),
            ],
        )
        LOG.info(f"gRPC channel created successfully using transport mode: {transport_mode}")
        return channel
    except Exception as e:
        LOG.error(f"Failed to create gRPC channel with mode '{transport_mode}': {e}")
        raise


def _wait_for_sherlock_grpc_ready(channel, timeout):
    # Check that the gRPC connection is up.
    try:
        LOG.info("Waiting for Sherlock gRPC service to start...")
        grpc.channel_ready_future(channel).result(timeout)
    except grpc.FutureTimeoutError:
        raise SherlockConnectionError(message="Error starting gRPC service")


def _get_base_ansys(
    year: Optional[int] = None, release_number: Optional[int] = None
) -> tuple[str, int]:
    supported_installed_versions = {
        env_key: path
        for env_key, path in os.environ.items()
        if env_key.startswith("AWP_ROOT") and os.path.isfile(_get_sherlock_exe_file_path(path))
    }

    sorted_installed_version_keys = sorted(supported_installed_versions, reverse=True)

    if year is not None:
        two_digit_year = _extract_sherlock_version_year(year)
        if release_number is None:
            # If no release number is provided, find the latest version for the given year
            for key in sorted_installed_version_keys:
                ansys_version = _get_ansys_version_from_awp_root(key)
                ansys_year = (int)(ansys_version / 10)
                if ansys_year == two_digit_year:
                    return supported_installed_versions[key], ansys_version
            raise ValueError(f"No versions of Sherlock are installed for year {year}.")
        else:
            try:
                sherlock_version = int(f"{two_digit_year}{release_number}")
                version_key = f"AWP_ROOT{sherlock_version}"
                if version_key in supported_installed_versions:
                    return supported_installed_versions[version_key], sherlock_version
                else:
                    raise ValueError(
                        f"Sherlock {two_digit_year} {release_number} is not installed."
                    )
            except ValueError as e:
                LOG.error(f"Error extracting Sherlock version year: {e}")
                raise e
    else:
        for key in sorted_installed_version_keys:
            ansys_version = _get_ansys_version_from_awp_root(key)
            if ansys_version >= _EARLIEST_SUPPORTED_VERSION:
                return supported_installed_versions[key], ansys_version

    raise ValueError("Could not find any installed version of Sherlock.")


def _get_ansys_version_from_awp_root(awp_root: str) -> int:
    if awp_root.find("AWP_ROOT") >= 0:
        return int(awp_root.replace("AWP_ROOT", ""))

    return 0


def _get_sherlock_exe_path(
    year: Optional[int] = None, release_number: Optional[int] = None
) -> tuple[str, int, str]:
    ansys_base, sherlock_version = _get_base_ansys(year=year, release_number=release_number)
    if not ansys_base:
        return "", 0, ""
    sherlock_bin = _get_sherlock_exe_file_path(ansys_base)
    return sherlock_bin, sherlock_version, ansys_base


def _get_sherlock_exe_file_path(ansys_base):
    if os.name == "nt":
        return os.path.join(ansys_base, "sherlock", "SherlockClient.exe")
    return os.path.join(ansys_base, "sherlock", "runSherlock")


def _extract_sherlock_version_year(year: int) -> int:
    if 1000 <= year <= 9999:
        return year % 100
    raise ValueError("Year must be a 4-digit integer.")
